#ifndef ARRAY_HPP_
#define ARRAY_HPP_

#include <vector>
#include <array>
#include <type_traits>
#include <initializer_list>
#include <stdexcept>

constexpr size_t Dynamic = SIZE_MAX;

/*! Array dimensions and strides: known at compile-time & dynamic
 */
template <size_t...>
struct Shape;

template <>
struct Shape<>
{
    static const size_t head = 1;
    static const size_t ndim = 0;
    static const bool fixed = true;
    static constexpr size_t dim(size_t i) { return 0; }
    static constexpr size_t stride(size_t i) { return 1; }
};

template <size_t Dim0, size_t... Dims>
struct Shape<Dim0,Dims...>
{
    static const size_t head = Dim0;
    typedef Shape<Dims...> Tail;
    static const size_t ndim = 1 + sizeof...(Dims);
    static const bool fixed = (Dim0 != Dynamic) && Tail::fixed;

    static constexpr size_t dim(size_t i)
        {
            return (i == 0) ? Dim0 : Tail::dim(i-1);
        }

    static constexpr size_t stride(size_t i)
        {
            if (i == 0) {
                constexpr size_t tail_stride = Tail::stride(0);

                if (Tail::head == Dynamic || tail_stride == Dynamic)
                    return Dynamic;

                return Tail::head * tail_stride;
            } else {
                return Tail::stride(i-1);
            }
        }

    static constexpr std::array<size_t, ndim> dims() { return {Dim0, Dims...}; }
};

namespace detail
{
    /*! Extract Nth tail shape */
    template <size_t I, typename Shape>
    struct TailNth;

    template <typename Shape>
    struct TailNth<0, Shape>
    {
        using type = Shape;
    };

    template <size_t I, typename Shape>
    struct TailNth
    {
        using type = typename TailNth<I-1, typename Shape::Tail>::type;
    };
}

/*! Simple data-by-reference strided array class a la Fortran
 */
template <typename Scalar, typename Shape, typename Storage = std::vector<Scalar> >
class Array
{
private:
    Storage& data_;
    std::array<size_t, Shape::ndim> shape_;
    std::array<size_t, Shape::ndim> stride_;
    size_t offset_ = 0;

    void data_bounds_check() const
        {
#ifndef NO_BOUNDS_CHECK
            size_t total = 1;

            for (size_t i = 0; i < Shape::ndim; ++i) {
                total *= shape_[i];
                if ((Shape::dim(i) != Dynamic && Shape::dim(i) != shape_[i])
                    || shape_[i] == Dynamic)
                    throw std::out_of_range("mismatch with fixed shape");
            }
            total += offset_;
            if (total > data_.size())
                throw std::out_of_range("data array too small");
#endif
        }

    void init_strides()
        {
            for (size_t i = Shape::ndim; i > 0; --i) {
                if (Shape::stride(i-1) != Dynamic)
                    stride_[i-1] = Shape::stride(i-1);
                else if (i == Shape::ndim)
                    stride_[i-1] = 1;
                else
                    stride_[i-1] = shape_[i] * stride_[i];
            }
        }

    template <size_t axis>
    static constexpr Eigen::Index eigen_shape()
        {
            size_t n = (axis == 0) ? Shape::head : Shape::Tail::head;
            return (n == Dynamic) ? Eigen::Dynamic : n;
        }

public:
    Array(Storage& data, const std::array<size_t, Shape::ndim> shape, const size_t offset=0)
        : data_(data), shape_(shape), offset_(offset)
        {
            init_strides();
            data_bounds_check();
        }

    Array(Storage& data)
        : data_(data), offset_(0)
        {
            static_assert(Shape::fixed, "array shape is not compile-time fixed");
            shape_ = Shape::dims();
            init_strides();
            data_bounds_check();
        }

    template <typename... Idx>
    Scalar& operator()(Idx... idxs) { return data_[index(idxs...)]; }

    template <typename... Idx>
    const Scalar&& operator()(Idx... idxs) const { return data_[index(idxs...)]; }

    Storage& data() { return data_; }

    template <size_t axis>
    const size_t shape() const
        {
            static_assert(axis < Shape::ndim, "invalid axis");
            constexpr size_t n = Shape::dim(axis);
            if (n != Dynamic)
                return n;
            return shape_[axis];
        }

    template <size_t axis>
    const size_t stride() const
        {
            static_assert(axis < Shape::ndim, "invalid axis");
            if (Shape::stride(axis) != Dynamic)
                return Shape::stride(axis);
            return stride_[axis];
        }

    size_t offset() const { return offset_; }

    template <typename... Idx>
    size_t index(Idx... idxs) const
        {
            static_assert(sizeof...(idxs) == Shape::ndim,
                          "number of indices must equal the number of dimensions");

            const std::array<size_t, Shape::ndim> m{idxs...};
            size_t idx = offset_;

            for (size_t i = 0; i < Shape::ndim; ++i) {
                if (Shape::stride(i) != Dynamic) {
                    idx += Shape::stride(i) * m[i];
                } else {
                    idx += stride_[i] * m[i];
                }
#ifndef NO_BOUNDS_CHECK
                if (m[i] >= shape_[i])
                    throw std::out_of_range("index out of bounds");
#endif
            }

#ifndef NO_BOUNDS_CHECK
            if (idx >= data_.size())
                throw std::out_of_range("total index out of bounds");
#endif
            return idx;
        }

    template <typename... Idx>
    Array<Scalar, typename detail::TailNth<sizeof...(Idx), Shape>::type > part(Idx... idxs) const
        {
            constexpr size_t nidxs = sizeof...(idxs);
            static_assert(nidxs < Shape::ndim,
                          "number of indices must be small than the number of dimensions");

            const std::array<size_t, Shape::ndim> m{idxs...};
            std::array<size_t, Shape::ndim - nidxs> shape;
            size_t offset = offset_;

            for (size_t i = 0; i < nidxs; ++i) {
                if (Shape::stride(i) != Dynamic) {
                    offset += Shape::stride(i) * m[i];
                } else {
                    offset += stride_[i] * m[i];
                }
#ifndef NO_BOUNDS_CHECK
                if (m[i] >= shape_[i])
                    throw std::out_of_range("index out of bounds");
#endif
            }

            for (size_t i = nidxs; i < Shape::ndim; ++i)
                shape[i - nidxs] = shape_[i];
            return {data_, shape, offset};
        }

    typedef Eigen::Matrix<Scalar, eigen_shape<0>(), eigen_shape<1>(), Eigen::RowMajor> EigenMatrix;
    typedef Eigen::Map<EigenMatrix> EigenMap;

    EigenMap matrix() const
        {
            static_assert(Shape::ndim == 2, "matrix must be two-dimensional");
            return {(Scalar *)data_.data() + offset_,
                static_cast<Eigen::Index>(shape<0>()), static_cast<Eigen::Index>(shape<1>())};
        }

    operator EigenMap() const { return matrix(); }
};


#endif
